EPS = 1e-6

# Values from Structure quality and target parameters (Engh and Huber)

# Backbone bond lengths
BL_N_CA = 1.459
BL_CA_C = 1.525
BL_C_N = 1.336
BL_C_O = 1.229

# Backbone bond angles
BA_N_CA_C = 111.0
BA_N_CA_CB = 110.6
BA_CA_C_N = 117.2
BA_CA_C_O = 120.1
BA_O_C_N = 122.7
BA_C_CA_CB = 110.6
BA_C_N_CA = 121.7

# Van der Waals radii
VDW_N = 1.55
VDW_C = 1.7